{-# OPTIONS -XRank2Types #-}

-- | The codensity monad transformer @Codensity@ adds support 
-- for computations in an effectful context @m@.

module Monatron.Codensity (
 Codensity,
 codensity,
 runCodensity
) where

import Monatron.MonadT
import Control.Monad.Fix

----------------------------------------------------------
-- Codensity Monad
----------------------------------------------------------

newtype Codensity f a = Codensity { 
      unCodensity :: forall b. (a -> f b) -> f b 
}

codensity :: (forall b. (a -> f b) -> f b) -> Codensity f a
codensity = Codensity

-- | Runs the computation in a trivial context.
runCodensity :: Monad m => Codensity m a -> m a
runCodensity c = unCodensity c return 

instance MonadT Codensity where
    lift m        = Codensity (m >>=)
    c `tbind` f   = Codensity (\k -> unCodensity c (\a -> unCodensity (f a) k))
    tmixmap f g m = Codensity $ \k -> f (unCodensity m (g . k))

instance (Monad m) => Monad (Codensity m) where
    return = treturn
    fail   = lift . fail
    (>>=)  = tbind

instance (Monad m) => Functor (Codensity m) where fmap = liftM

-- still need to prove that MonadFix laws hold
instance MonadFix m => MonadFix (Codensity m) where
    mfix f = Codensity $ \k -> mfix (runCodensity. f) >>= k

------------------------

